/*
 * Copyright (C) 2006 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

/**
 * \file grab_client.c functions for grabbing the stream at any position
 * in a filter chain
 *
 * \sa filter_chain filter_chain_info filter
 */

#include "para.h"
#include "grab_client.cmdline.h"
#include "list.h"
#include "sched.h"
#include "filter.h"
#include "grab_client.h"
#include "audiod.h"
#include "error.h"
#include "string.h"
#include "fd.h"

/** grab clients that are not yet attached to a filter node */
struct list_head inactive_grab_client_list;

static int max_num_filters(void)
{
	int i, ret = 0;
	for (i = 0; audio_formats[i]; i++) {
		PARA_INFO_LOG("%s filter chain length: %d\n", audio_formats[i],
			num_filters(i));
		ret = PARA_MAX(ret, num_filters(i));
	}
	PARA_INFO_LOG("maximal filter chain length: %d\n", ret);
	return ret;
}

static int gc_write(char *buf, size_t len, struct filter_callback *fcb)
{
	struct grab_client *gc = fcb->data;
	struct timeval tv = {0, 100};
	int ret;

//	PARA_INFO_LOG("writing %d bytes to fd %d\n", len, gc->fd);
	fd_set wfds;
	FD_ZERO(&wfds);
	FD_SET(gc->fd, &wfds);
	ret = para_select(gc->fd + 1, NULL, &wfds, &tv);
	if (ret <= 0) {
		if (gc->mode == GRAB_PEDANTIC)
			return -E_PEDANTIC_GRAB;
		if (gc->mode == GRAB_SLOPPY)
			return 1;
	}
rewrite:
	ret = write(gc->fd, buf, len);
	if (ret < 0) {
		ret = -E_GC_WRITE;
		gc->error = E_GC_WRITE;
	} else {
		if (ret != len) {
			if (gc->mode == GRAB_PEDANTIC)
				return -E_PEDANTIC_GRAB;
			if (gc->mode == GRAB_AGGRESSIVE) {
				len -= ret;
				memmove(buf, buf + ret, len);
				goto rewrite;
			}
		}
	}
	return ret;
}

/* TODO: gengetopt can handle the grab client modes */
static int check_gc_args(struct grab_client *gc)
{
	int i;
	struct grab_client_args_info *c = gc->conf;
	char **mv = grab_client_cmdline_parser_mode_values;

	PARA_INFO_LOG("filter_num: %d\n", c->filter_num_arg);
	for (i = 0; mv[i]; i++)
		if (!strcmp(c->mode_arg, mv[i]))
			break;
	if (!mv[i])
		return -E_GC_SYNTAX;
	gc->mode = i;
	gc->audio_format_num = -1;
	if (c->audio_format_given) {
		gc->audio_format_num = get_audio_format_num(c->audio_format_arg);
		if (gc->audio_format_num < 0)
			return gc->audio_format_num;
	}
	if (c->slot_arg > MAX_STREAM_SLOTS)
		return -E_BAD_GC_SLOT;
	if (c->filter_num_arg <= 0)
		return -E_BAD_GC_FILTER_NUM;
	if (c->audio_format_given) {
		if (num_filters(gc->audio_format_num) < c->filter_num_arg)
			return -E_BAD_GC_FILTER_NUM;
	} else
		if (c->filter_num_arg > max_num_filters())
			return -E_BAD_GC_FILTER_NUM;

	return 1;
}

static void add_inactive_gc(struct grab_client *gc)
{
	PARA_INFO_LOG("adding grab client %p (fd %d) to inactive list\n",
		gc, gc->fd);
	list_add(&gc->node, &inactive_grab_client_list);
}

static void gc_free(struct grab_client *gc)
{
	int i;

	for (i = 0; i < gc->argc; i++)
		free(gc->argv[i]);
	free(gc->argv);
	free(gc->conf);
	free(gc);

}

static void gc_close(struct filter_callback *fcb)
{
	struct grab_client *gc = fcb->data;

	if (gc->conf->one_shot_given || gc->error) {
		PARA_INFO_LOG("closing fd %d (grab client %p)\n", gc->fd, gc);
		close(gc->fd);
		gc_free(gc);
		/* close on fork ?*/
		return;
	}
	add_inactive_gc(gc);
}

/**
 * move a grab client from the inactive list to a filter node
 *
 * \param gc the grab client to activate
 * \param fn the filter node \a gc gets attached to
 *
 * \sa filter_node::callbacks, inactive_grab_client_list
 */
void activate_grab_client(struct grab_client *gc, struct filter_node *fn)
{
	PARA_INFO_LOG("activating %p (fd %d, filter node: %p)\n", gc, gc->fd, fn);
	list_del(&gc->node);
	list_add(&gc->fcb.node, &fn->callbacks);
}

/**
 * activate inactive grab clients if possible
 *
 * \param slot_num audiod's slot for the new audio file
 * \param audio_format_num the number of the audio format of the new audio file
 * \param filter_list the list of activated filters for that new audio file
 *
 * This is called from audiod.c when the current audio file changes. It loops
 * over all inactive grab clients and checks each grab client's configuration
 * to determine if the client in question wishes to grab the new stream.  If
 * yes, this grab client is moved from the inactive grab client list to an
 * appropriate filter_node.
 *
 * \sa filter_chain_info::filters, inactive_grab_client_list,
 * activate_grab_client
 */
void activate_inactive_grab_clients(int slot_num, int audio_format_num,
		struct list_head *filter_list)
{
	struct grab_client *gc, *tmp;
	int i;
	struct filter_node *fn;

	list_for_each_entry_safe(gc, tmp, &inactive_grab_client_list, node) {
//		PARA_INFO_LOG("checking inactive grab client %p\n", gc);
		if (gc->conf->slot_arg >= 0 && gc->conf->slot_arg != slot_num)
			continue;
		if (gc->audio_format_num >= 0 && gc->audio_format_num !=
				audio_format_num)
			continue;
		if (gc->conf->filter_num_arg >= 0 &&
				num_filters(gc->audio_format_num)
				< gc->conf->filter_num_arg)
			continue;
		i = 1;
		list_for_each_entry(fn, filter_list, node) {
			if (gc->conf->filter_num_arg <= 0
				|| i == gc->conf->filter_num_arg)
				break;
			i++;
		}
		activate_grab_client(gc, fn);
	}
}

/**
 * check the command line options and allocate a grab_client structure
 *
 * \param fd the file descriptor of the client
 * \param line the command line
 * \param err non-zero if an error occured
 *
 * If the command line options given by \a argc and \a argv are valid.
 * allocate a struct grab_client and initialize it with this valid
 * configuration. Moreover, add the new grab client to the inactive list.
 *
 * \return On success, this function returns a pointer to the newly created
 * struct. On errors, it returns NULL and sets \a err appropriately.
 *
 * \sa grab_client, inactive_grab_client_list, activate_grab_client,
 * filter_node::callbacks
 */
/*
 * argc, argv get freed when com_grab() returns, so we have to make a
 * copy.
 */
struct grab_client *grab_client_new(int fd, char *line, int *err)
{
	int ret;
	struct grab_client *gc = para_calloc(sizeof(struct grab_client));

	gc->conf = para_calloc(sizeof(struct grab_client_args_info));

	ret = grab_client_cmdline_parser_string(line, gc->conf, "grab");
	*err = -E_GC_SYNTAX;
	if (ret)
		goto err_out;
	*err = -E_GC_HELP_GIVEN;
	if (gc->conf->help_given)
		goto err_out;
	*err = -E_GC_VERSION_GIVEN;
	if (gc->conf->version_given)
		goto err_out;
	*err = check_gc_args(gc);
	if (*err < 0)
		goto err_out;
	if (gc->conf->input_grab_given) {
		gc->fcb.input_cb = gc_write;
		gc->fcb.output_cb = NULL;
	} else {
		gc->fcb.output_cb = gc_write;
		gc->fcb.input_cb = NULL;
	}
	gc->fd = fd;
	gc->fcb.close = gc_close;
	gc->fcb.data = gc;
	add_inactive_gc(gc);
	return gc;
err_out:
	free(gc->conf);
	free(gc);
	return NULL;
}

/** initialize the grabbing subsystem.
 *
 * This has to be called once during startup before any other function from
 * grab_client.c may be used. It initializes \a inactive_grab_client_list.
 */
void init_grabbing()
{
	PARA_INFO_LOG("%s", "grab init\n");
	INIT_LIST_HEAD(&inactive_grab_client_list);
}

